"""Test Dataset."""

from __future__ import annotations

from .common import DATADIR, load_txt
from .dataset import Dataset, LabelledDataset


def load_test_dataset(lines: bool = False) -> Dataset:
    """Load the very simple test dataset as :class:`Dataset`.

    Args:
        lines (bool, optional): consider each line as a document rather than each
            file. Defaults to False.

    Returns:
        Dataset: the very simple test dataset
    """
    _write_test_dataset()
    return load_txt(DATADIR / "Test", lines=lines)


def load_labelled_test_dataset(lines: bool = False) -> LabelledDataset:
    """Load the very simple test dataset as :class:`LabelledDataset`.

    Args:
        lines (bool, optional): consider each line as a document rather than each
            file. Defaults to False.

    Returns:
        LabelledDataset: the very simple test dataset with labels
    """
    _write_test_dataset()
    return load_txt(DATADIR / "Test", lines=lines, label=True)


def _write_test_dataset(force: bool = False) -> None:
    """Write a simple dataset as collection of text file in the DATA directory.

    Args:
        force (bool, optional): force the rewriting of the dataset even if the dataset
            folder already exists. Defaults to False.
    """
    testdir = DATADIR / "Test"
    if (not testdir.exists()) or force:
        testdir.mkdir(exist_ok=True)

        animals = testdir / "animals"
        animals.mkdir(exist_ok=True)
        (animals / "dinosaur.txt").write_text(
            (
                "Dinosaurs are reptiles.\n"
                "Birds descend from dinosaurs.\n"
                "Even if most of the dinosaurs don't fly, "
                "probably the majority of them was covered by feathers."
            )
        )
        (animals / "bird.txt").write_text(
            (
                "Birds lay eggs, like reptiles and fishes.\n"
                "Most of the birds fly, even if penguins, "
                "ostriches and some others are unable to fly.\n"
                "Birds have two wings, like some dinosaurs and others ancient reptiles."
            )
        )

        maths = testdir / "maths"
        maths.mkdir(exist_ok=True)
        (maths / "geometry.txt").write_text(
            (
                "Geometry studies shapes and entities in the space.\n"
                "A geometrician proves theorem about the relation among two "
                "or more geometrical entities.\n"
                "Continuity is a geometric concept widely used in calculus."
            )
        )
        (maths / "linearalgebra.txt").write_text(
            (
                "Linear algebra studies matrices, vectors and vectorial spaces.\n"
                "Sometimes linear algebra is considered a subfield of geometry.\n"
                "Many theorem regarding matrices exists but one of "
                "the most important is the spectral one."
            )
        )



"""Common variable and function for adso module."""

import os
import random
from pathlib import Path

import numpy as np

ADSODIR = os.getenv("ADSODIR")
if ADSODIR is not None:
    ADSODIR = Path(ADSODIR)
else:
    ADSODIR = Path.home() / ".adso"


def set_seed(seed: int) -> None:
    """Set random seed for reproducibility.

    Take care of python random library and numpy.

    Args:
        seed (int): the value choosen as seed for the random generators
    """
    random.seed(seed)
    np.random.seed(seed)
    # torch.manual_seed(seed)



"""20newgroups datatset.

Download the 20newsgroup dataset from `Jason Rennie's site
    <http://qwone.com/~jason/20Newsgroups/>` and import it as :class:`LabelledDataset`
"""

import tarfile
from typing import Tuple, Union

import requests

from .common import DATADIR, load_txt
from .dataset import LabelledDataset


def load_20newsgroups(
    split: str = "all",
) -> Union[LabelledDataset, Tuple[LabelledDataset, LabelledDataset]]:
    """Load the 20newsgroups dataset with labels from
    <http://qwone.com/~jason/20Newsgroups/>`

    Args:
        split (str, optional): which part of the dataset have to be load:

            * "all" for a single :class:`LabelledDataset` with all the available
              documents
            * "train" for the train split only (deterministic)
            * "test" for the test split only (deterministic)
            * "both" for the tuple (train, test)

            Defaults to "all".

    Returns:
        Union[LabelledDataset, Tuple[LabelledDataset, LabelledDataset]]:
            the 20newsgroups dataset with labels, as LabelledDataset
    """
    newsdir = DATADIR / "20newsgroups"
    newsdir.mkdir(parents=True, exist_ok=True)
    if not any(newsdir.iterdir()):
        data = requests.get(
            "http://qwone.com/~jason/20Newsgroups/20news-bydate.tar.gz",
            allow_redirects=True,
            stream=True,
        )
        with (newsdir / "20news.tar.gz").open("wb") as file:
            file.write(data.content)
        tarfile.open(name=(newsdir / "20news.tar.gz"), mode="r|gz").extractall(
            path=newsdir
        )

    train = load_txt(newsdir / "20news-bydate-test", label=True, extension=None)
    test = load_txt(newsdir / "20news-bydate-test", label=True, extension=None)

    if split == "both":
        return train, test
    elif split == "train":
        return train
    elif split == "test":
        return test
    elif split == "all":
        return train + test
    else:
        raise ValueError("Split can be both, train, test or all")



"""Adso is a topic-modelling library based on numpy and nltk."""

from . import data, topicmodel, transform
from .common import ADSODIR, set_seed

# Create adso folder
ADSODIR.mkdir(exist_ok=True, parents=True)



"""Vectorizer class."""

from __future__ import annotations

import abc
from collections import Counter
from functools import reduce
from itertools import chain, starmap
from typing import Dict, List, Union

import numpy as np

import scipy as sp

from .common import Transformer


class Vocab:  # Inspiraed by torchtext.vocab.Vocab
    """Vocabulary class, used to convert string to int and viceversa."""

    def __init__(
        self: Vocab,
        count: Counter,
        max_size: Union[None, int] = None,
        min_freq: float = 0,
        max_freq: float = 1,
        min_count: int = 1,
    ) -> None:
        """Construct the vocabulary from a :class:`collections.Counter` object with
        word count.

        Args:
            count (Counter): a :class:`collections.Counter` which store the list of
                words and their count, with which build the vocabulary.
            max_size (Union[None, int], optional): maximum number of word to store into
                the vocabulary. Most common words are kept.
                None to keep all words. Defaults to None.
            min_freq (float, optional): minimum frequency (as count of word / sum of
                counts of all words) to keep a word. Defaults to 0.
            max_freq (float, optional): maximum frequency (as count of word / sum of
                counts of all words) to keep a word. Defaults to 1.
            min_count (int, optional): minimum number of occurency (i.e. associated
                value in Counter object) to keep a word. Defaults to 1.

        Attributes:
            itos (Dict[int, str]): dictionary to access words in vocabulary
                given the index.
            stoi (Dict[str, int]): dictionary to access the index of a given
                word in vocabulary.
        """
        self.min_freq = min_freq
        self.max_freq = max_freq
        self.min_count = min_count

        total_count = sum(count.values())
        min_freq = max(self.min_freq * total_count, self.min_count)
        max_freq = self.min_freq * total_count

        words = dict(
            Counter(
                dict(
                    filter(
                        lambda x: (x[1] >= max_freq) or (x[1] <= min_freq),
                        count.items(),
                    )
                )
            ).most_common(max_size)
        ).keys()

        self.itos: Dict[int, str] = dict(enumerate(words))
        self.stoi: Dict[str, int] = dict(map(reversed, self.itos.items()))

    def __getitem__(self: Vocab, index: Union[int, str]) -> Union[str, int]:
        """Enable [ ] syntax for both integer (return correspondent word) and strings
        (returning correspondent index).

        Args:
            index (Union[int, str]): a word to get the correspondent inde or an index
                to get the correspondent word.

        Returns:
            Union[str, int]: the word or the index correspondent to the index or
                the word selected.
        """
        if isinstance(index, str):
            return self.stoi[index]
        elif isinstance(index, int):
            return self.itos[index]
        else:
            raise ValueError("index must be a string or an int")

    def __len__(self: Vocab) -> int:
        """Define len( ) for Vocab class.

        Returns:
            int: number of words in the vocabulary.
        """
        return len(self.stoi)


class Vectorizer(Transformer):
    """Abstract class to group vectorizer, i.e. text to number processors."""

    @abc.abstractmethod
    def fit(data: List[List[str]]) -> None:
        """Create the vocabulary.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.
        """
        pass

    @abc.abstractmethod
    def transform(data: List[List[str]]) -> Union[np.array, sp.sparse.spmatrix]:
        """Trasform the collection of tokens in a sequence of numerical values,
        given an already fitted vocabulary.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.

        Returns:
            Union[np.array, sp.sparse.spmatrix]: a matrix which maps documents to the
                rows.
        """
        pass

    @abc.abstractmethod
    def fit_transform(data: List[List[str]]) -> Union[np.array, sp.sparse.spmatrix]:
        """Trasform the collection of tokens in a sequence of numerical values,
        creating a vocabulary in the process.

         Args:
             data (List[List[str]]): a list of list of tokens. Outer list map to
                 documents while inner list map to words in each document. For example
                 the output of a :class:`transform.Tokenizer`.

         Returns:
             Union[np.array, sp.sparse.spmatrix]: a matrix which maps documents to the
                 rows.
        """
        pass


class CountVectorizer(Vectorizer):
    """Create a document-term matrix with count as values."""

    def __init__(
        self: CountVectorizer,
        max_size: Union[None, int] = None,
        min_freq: float = 0,
        max_freq: float = 1,
        min_count: int = 1,
    ) -> None:
        """Initialize the vectorizer and set the parameters.

        Args:
            max_size (Union[None, int], optional): maximum number of word to store into
                the vocabulary. Most common words are kept.
                None to keep all words. Defaults to None.
            min_freq (float, optional): minimum frequency (as count of word / sum of
                counts of all words) to keep a word. Defaults to 0.
            max_freq (float, optional): maximum frequency (as count of word / sum of
                counts of all words) to keep a word. Defaults to 1.
            min_count (int, optional): minimum number of occurency to keep a word.
                Defaults to 1.

        Attributes:
            vocab: the stored vocabulary used to transform the data.
        """
        if not ((0 <= min_freq <= 1) and (0 <= max_freq <= 1)):
            raise ValueError("min_freq and max_freq must be in the [0,1] interval")

        self.vocab: Vocab
        self.max_size = max_size
        self.min_freq = min_freq
        self.max_freq = max_freq
        self.min_count = min_count

    def fit(
        self: CountVectorizer,
        data: List[List[str]],
    ) -> None:
        """Create the vocabulary.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.
        """
        count = reduce(lambda x, y: x + y, map(Counter, data))
        self.vocab = Vocab(
            count, self.max_size, self.min_freq, self.max_freq, self.min_count
        )

    def transform(self: CountVectorizer, data: List[List[str]]) -> sp.sparse.spmatrix:
        """Return a document-term matrix with count as values. Words not in vocabulary
        are discarded.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.

        Returns:
            sp.sparse.spmatrix: a document-term matrix with count as values.
        """
        if not self.vocab:
            NameError("It is necessary to fit before transform")

        def toindex(s: str) -> Union[None, str]:
            try:
                return self.vocab[s]
            except KeyError:
                return None

        def document_to_count(document: List[str]):
            return list(
                Counter(
                    list(
                        filter(
                            lambda tkn: tkn is not None,
                            map(toindex, document),
                        )
                    )
                ).items()
            )

        coo_sparse = np.array(
            list(
                chain.from_iterable(
                    starmap(
                        lambda row, doc: list(
                            starmap(lambda col, val: (val, row, col), doc)
                        ),
                        enumerate(map(document_to_count, data)),
                    )
                )
            )
        )

        return sp.sparse.csr_matrix(
            (coo_sparse[:, 0], (coo_sparse[:, 1], coo_sparse[:, 2])),
            shape=(len(data), len(self.vocab)),
            dtype=int,
        )

    def fit_transform(
        self: CountVectorizer,
        data: List[List[str]],
    ) -> Union[np.array, sp.sparse.spmatrix]:
        """Return a document-term matrix with count as values. Vocabulary is built
        on the data passed as input.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.

        Returns:
            sp.sparse.spmatrix: a document-term matrix with count as values.
        """
        self.fit(data)
        return self.transform(data)


class ListVectorizer(CountVectorizer):
    """Create a matrix with the ordered list of indeces for each document
    on the rows."""

    def __init__(
        self: ListVectorizer,
        max_size: Union[None, int] = None,
        min_freq: float = 0,
        max_freq: float = 1,
        min_count: int = 1,
        max_length: int = 100,
    ) -> None:
        """Initialize the vectorizer and set the parameters.

        Args:
            max_size (Union[None, int], optional): maximum number of word to store into
                the vocabulary. Most common words are kept.
                None to keep all words. Defaults to None.
            min_freq (float, optional): minimum frequency (as count of word / sum of
                counts of all words) to keep a word. Defaults to 0.
            max_freq (float, optional): maximum frequency (as count of word / sum of
                counts of all words) to keep a word. Defaults to 1.
            min_count (int, optional): minimum number of occurency to keep a word.
                Defaults to 1.
            max_length (int, optional): maximum lenght of each row, i.e. maximum number
                of words for each document to keep. The first N words are kept.
                Defaults to 100.

        Attributes:
            vocab: the stored vocabulary used to transform the data.
        """
        if not ((0 <= min_freq <= 1) and (0 <= max_freq <= 1)):
            raise ValueError("min_freq and max_freq must be in the [0,1] interval")

        super().__init__(
            max_size=max_size, min_freq=min_freq, max_freq=max_freq, min_count=min_count
        )
        self.max_length = max_length

    def fit(
        self: ListVectorizer,
        data: List[List[str]],
    ) -> None:
        """Create the vocabulary.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.
        """
        super().fit(data)

    def transform(self: ListVectorizer, data: List[List[str]]) -> np.array:
        """Return a matrix with the ordered list of indeces for each document
        on the rows. Words not in vocabulary are discarded.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.

        Returns:
            np.array: a matrix with the ordered list of indeces for each document
            on the rows.
        """
        if not self.vocab:
            NameError("It is necessary to fit before transform")

        def toindex(s: str) -> Union[None, str]:
            try:
                return self.vocab[s]
            except KeyError:
                return None

        def document_to_list(document: list(str)):
            return list(
                filter(
                    lambda tkn: tkn is not None,
                    map(toindex, document),
                )
            )

        data = list(map(document_to_list, data))
        length = min(max(map(len, data)), self.max_length)

        def pad(data: list, length: int):
            if len(data) > length:
                return data[:length]
            else:
                return data + [np.nan] * (length - len(data))

        return np.array(list(map(lambda row: pad(row, length), data)))

    def fit_transform(
        self: ListVectorizer,
        data: List[List[str]],
    ) -> np.array:
        """Return a matrix with the ordered list of indeces for each document
        on the rows. Vocabulary is built on the data passed as input.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.

        Returns:
            np.array: a matrix with the ordered list of indeces for each document
            on the rows.
        """
        self.fit(data)
        return self.transform(data)


class FreqVectorizer(CountVectorizer):
    """Create a document-term matrix with frequencies as values."""

    def __init__(
        self: FreqVectorizer,
        max_size: Union[None, int] = None,
        min_freq: float = 0,
        max_freq: float = 1,
        min_count: int = 1,
    ) -> None:
        """Initialize the vectorizer and set the parameters.

        Args:
            max_size (Union[None, int], optional): maximum number of word to store into
                the vocabulary. Most common words are kept.
                None to keep all words. Defaults to None.
            min_freq (float, optional): minimum frequency (as count of word / sum of
                counts of all words) to keep a word. Defaults to 0.
            max_freq (float, optional): maximum frequency (as count of word / sum of
                counts of all words) to keep a word. Defaults to 1.
            min_count (int, optional): minimum number of occurency to keep a word.
                Defaults to 1.

        Attributes:
            vocab: the stored vocabulary used to transform the data.
        """
        super().__init__(
            max_size=max_size, min_freq=min_freq, max_freq=max_freq, min_count=min_count
        )

    def fit(
        self: FreqVectorizer,
        data: List[List[str]],
    ) -> None:
        """Create the vocabulary.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.
        """
        super().fit(data)

    def transform(self: FreqVectorizer, data: List[List[str]]) -> sp.sparse.spmatrix:
        """Return a document-term matrix with frequencies as values. Words not in
        vocabulary are discarded.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.

        Returns:
            sp.sparse.spmatrix: a document-term matrix with frequencies as values.
        """
        matrix = super().transform(data)
        # row normalization
        return sp.sparse.diags(1 / matrix.sum(axis=1).A1) @ matrix

    def fit_transform(
        self: FreqVectorizer,
        data: List[List[str]],
    ) -> sp.sparse.spmatrix:
        """Return a document-term matrix with frequencies as values. Vocabulary is built
        on the data passed as input.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.

        Returns:
            sp.sparse.spmatrix: a document-term matrix with frequencies as values.
        """
        self.fit(data)
        return self.transform(data)


class TFIDFVectorizer(CountVectorizer):
    """Create a document-term matrix with TFIDF frequencies as values."""

    def __init__(
        self: TFIDFVectorizer,
        max_size: Union[None, int] = None,
        min_freq: float = 0,
        max_freq: float = 1,
        min_count: int = 1,
        smooth: bool = False,
        log_df: bool = False,
    ) -> None:
        """[summary]

        Args:
            self (TFIDFVectorizer): [description]
            max_size (Union[None, int], optional): [description]. Defaults to None.
            min_freq (float, optional): [description]. Defaults to 0.
            max_freq (float, optional): [description]. Defaults to 1.
            smooth (bool, optional): [description]. Defaults to False.
            log_df (bool, optional): [description]. Defaults to False.
        """
        super().__init__(
            max_size=max_size, min_freq=min_freq, max_freq=max_freq, min_count=min_count
        )
        self.smooth = smooth
        self.log_df = log_df

    def fit(
        self: TFIDFVectorizer,
        data: List[List[str]],
    ) -> None:
        """Create the vocabulary.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.
        """
        super().fit(data)

    def transform(self: TFIDFVectorizer, data: List[List[str]]) -> sp.sparse.spmatrix:
        """Return a document-term matrix with TFIDF frequencies as values. Words not in
        vocabulary are discarded.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.

        Returns:
            sp.sparse.spmatrix: a document-term matrix with TFIDF frequencies
                as values.
        """
        matrix = super().transform(data)
        tf = sp.sparse.diags(1 / matrix.sum(axis=1).A1) @ matrix
        delta = 1 if self.smooth else 0
        if self.log_df:
            idf = np.log(matrix.shape[0] + delta) - np.log(
                (matrix > 0).astype(int).sum(axis=0).A1 + delta
            )
        else:  # linear tfidf
            idf = (matrix.shape[0] + delta) / (matrix > 0).astype(int).sum(axis=0).A1
        return tf @ sp.sparse.diags(idf)

    def fit_transform(
        self: TFIDFVectorizer, data: List[List[str]]
    ) -> sp.sparse.spmatrix:
        """Return a document-term matrix with TFIDF frequencies as values. Vocabulary
        is built on the data passed as input.

        Args:
            data (List[List[str]]): a list of list of tokens. Outer list map to
                documents while inner list map to words in each document. For example
                the output of a :class:`transform.Tokenizer`.

        Returns:
            sp.sparse.spmatrix: a document-term matrix with TFIDF frequencies as
                values.
        """
        self.fit(data)
        return self.transform(data)



"""Trasform string tensor to numerical."""

import nltk

from .common import NLTKDIR, nltk_download
from .tokenizer import Tokenizer
from .vectorizer import (
    CountVectorizer,
    FreqVectorizer,
    ListVectorizer,
    TFIDFVectorizer,
    Vocab,
)

NLTKDIR.mkdir(exist_ok=True, parents=True)
nltk.data.path.append(NLTKDIR)



"""Non Negative Matrix Factorization.

AHCLS algorithm from Langville et al. ArXiv:1407.7299
"""

from __future__ import annotations

from typing import Tuple

import numpy as np

import scipy as sp

from .common import TopicModel


class NMF(TopicModel):
    """Nonnegative Matrix Factorization algorithm.

    Implemented following Langville et al. (ArXiv:1407.7299) using
    Alternate Least Squares.
    """

    def __init__(
        self: NMF,
        n_topic: int = 10,
        max_iter: int = 200,
        tolerance: float = 0,
        lambdaW: float = 0.5,
        lambdaH: float = 0.5,
        alphaW: float = 0.5,
        alphaH: float = 0.5,
        method: str = "ACLS",
    ) -> None:
        """Initialize NMF algorithm specifing parameters.

        Args:
            n_topic (int, optional): number of topic in corpus to estimate.
                Defaults to 10.
            max_iter (int, optional): maximum number of iteration for the algorithm.
                Defaults to 200.
            tolerance (float, optional): maximum relative error between iteration to
                early stop the algorithm. Defaults to 0.
            lambdaW (float, optional): smoothing parameter for W matrix
                (document-topic). Used in both ACLS and AHCLS variant.
                Must be positive. Defaults to 0.5.
            lambdaH (float, optional): smoothing parameter for H matrix (topic-word).
                Used in both ACLS and AHCLS variant. Must be positive. Defaults to 0.5.
            alphaW (float, optional): sparsity parameter for W matrix in AHCLS
                algorithm. Must be in (0,1). Defaults to 0.5.
            alphaH (float, optional): sparsity parameter for H matrix in AHCLS
                algorithm. Must be in (0,1). Defaults to 0.5.
            method (str, optional): the desired algoritm.
                One of ["ALS", "ACLS", "AHCLS"]. Defaults to "ACLS".
        """
        if n_topic > 0:
            self.n_topic = n_topic
        else:
            raise ValueError("n_topic must be positive")
        if max_iter > 0:
            self.max_iter = max_iter
        else:
            raise ValueError("max_iter must be positive")
        if lambdaW >= 0 and lambdaH >= 0:
            self.lambdaW = lambdaW
            self.lambdaH = lambdaH
        else:
            raise ValueError("lambda parameters must be non negative")
        if (1 >= alphaW >= 0) and (1 >= alphaH >= 0):
            self.alphaW = alphaW
            self.alphaH = alphaH
        else:
            raise ValueError("alpha parameters must be in [0,1]")
        if tolerance >= 0:
            self.tolerance = tolerance
        else:
            raise ValueError("tolerance must be nonnegative")
        if method in ["ALS", "ACLS", "AHCLS"]:
            self.method = method
        else:
            raise ValueError("Only ALS, ACLS, AHCLS methods are implemented")

    def fit(self: NMF, data: sp.sparse.spmatrix) -> sp.sparse.spmatrix:
        """Computate the H matrix (topic-document).

        Equivalent to call :func:`~topicmodel.NMF.fit_transform` method, which
        should be used instead.

        Args:
            data (sp.sparse.spmatrix): a document-term matrix for example the output
                of a subclass of :class:`transform.vectorizer.Vectorizer`.

        Returns:
            sp.sparse.spmatrix: topic-word H matrix.
        """
        self.fit_transform(data)
        return self.H

    def transform(self: NMF, data: sp.sparse.spmatrix) -> sp.sparse.spmatrix:
        """Estimate a topic-document matrix given the stored topic-word matrix.

        Args:
            data (sp.sparse.spmatrix): a document-term matrix for example the output
                of a subclass of :class:`transform.vectorizer.Vectorizer`.
                Must be of the same type of the one used to fit the model.

        Returns:
            sp.sparse.spmatrix: topic-document W matrix.
        """
        # solve Ht @ Wt = At equivalent to W @ H = A
        if data.shape[1] != self.H.shape[1]:
            raise ValueError(
                "shape of data is incompatible with the shape of the previuosly fitted topic-term matrix. The same vocab used for fitted data must be used for newer data"
            )
        return sp.sparse.linalg.spsolve(
            self.H.transpose(), data.transpose()
        ).transpose()

    def fit_transform(
        self: NMF, data: sp.sparse.spmatrix
    ) -> Tuple[sp.sparse.spmatrix, sp.sparse.spmatrix, int]:
        """Decompose a document-word matrix in two nonnegative document-topic
        and topic word matrix.

        Args:
            data (sp.sparse.spmatrix): a document-term matrix for example the output
                of a subclass of :class:`transform.vectorizer.Vectorizer`.

        Returns:
            Tuple[sp.sparse.spmatrix, sp.sparse.spmatrix, int]:
            document-topic W matrix, topic-word H matrix and the number of iterations.
        """
        A = data
        n_doc, n_term = data.shape
        n_topic = self.n_topic
        tolerance = self.tolerance

        if self.method == "AHCLS":
            betaH = ((1 - self.alphaH) * (n_topic ** 0.5) + self.alphaH) ** 2
            betaW = ((1 - self.alphaW) * (n_topic ** 0.5) + self.alphaW) ** 2

        if n_term / n_topic > 20:
            p = 20
        else:
            p = max(int(n_term / n_topic), 1)

        W = sp.sparse.csr_matrix(
            np.mean(
                np.resize(
                    data[
                        :, np.random.choice(n_term, size=p * n_topic, replace=False)
                    ].todense(),
                    (n_doc, n_topic, p),
                ),
                axis=2,
            )
        )

        if tolerance > 0:
            trAtA = (A.transpose() @ A).diagonal().sum()

            def error(
                H: sp.sparse.spmatrix, WtA: sp.sparse.spmatrix, WtW: sp.sparse.spmatrix
            ) -> float:
                return (
                    trAtA
                    - 2 * (H.transpose() @ WtA).diagonal().sum()
                    + (H.transpose() @ (WtW @ H)).diagonal().sum()
                ).item()

            err: float = n_doc * n_term

        else:

            def error(
                H: sp.sparse.spmatrix, WtA: sp.sparse.spmatrix, WtW: sp.sparse.spmatrix
            ) -> float:
                return 0

        if self.method == "AHCLS":

            def updateH(WtW: sp.sparse.matrix) -> sp.sparse.matrix:
                return sp.sparse.linalg.spsolve(
                    WtW.todense()
                    + self.lambdaH * betaH * sp.sparse.eye(n_topic)
                    - self.lambdaH,
                    WtA,
                )

            def updateW(H: sp.sparse.matrix) -> sp.sparse.matrix:
                return sp.sparse.linalg.spsolve(
                    (H @ H.transpose()).todense()
                    + self.lambdaW * betaW * sp.sparse.eye(n_topic)
                    - self.lambdaW,
                    H @ A.transpose(),
                ).transpose()

        elif self.method == "ACLS":

            def updateH(WtW: sp.sparse.matrix) -> sp.sparse.matrix:
                return sp.sparse.linalg.spsolve(
                    WtW - self.lambdaH * sp.sparse.eye(n_topic),
                    WtA,
                )

            def updateW(H: sp.sparse.matrix) -> sp.sparse.matrix:
                return sp.sparse.linalg.spsolve(
                    H @ H.transpose() - self.lambdaW * sp.sparse.eye(n_topic),
                    H @ A.transpose(),
                ).transpose()

        elif self.method == "ALS":

            def updateH(WtW: sp.sparse.matrix) -> sp.sparse.matrix:
                return sp.sparse.linalg.spsolve(
                    WtW,
                    WtA,
                )

            def updateW(H: sp.sparse.matrix) -> sp.sparse.matrix:
                return sp.sparse.linalg.spsolve(
                    H @ H.transpose(),
                    H @ A.transpose(),
                ).transpose()

        for iter in range(self.max_iter):
            WtW = W.transpose() @ W
            WtA = W.transpose() @ A
            H = updateH(WtW)
            H[H < 0] = 0
            H.eliminate_zeros()
            W = updateW(H)
            W[W < 0] = 0
            W.eliminate_zeros()

            if iter % 10 == 0:
                if tolerance > 0:
                    old_err = err
                    err = error(H, WtA, WtW)
                    print(f"Iteration {iter+1} - Error {err}")
                    if tolerance > abs(old_err - err) / old_err:
                        self.H = H
                        return W, H, iter + 1
                else:
                    print(f"Iteration {iter+1}")

        self.H = H
        return W, H, iter + 1



"""Common variable and function for data submodule."""

import abc

import nltk

from ..common import ADSODIR

NLTKDIR = ADSODIR / "NLTK"


def nltk_download(id: str):
    """Wrapper for :func:`nltk.downloader.download`, to download nltk data inside
    adso directory.

    Args:
        id (str): id of the required downloadable data. Cfr
            <http://www.nltk.org/nltk_data/>
    """
    return nltk.downloader.download(id, download_dir=NLTKDIR)


class Transformer(abc.ABC):
    """Abstract class for data manipulation class."""

    @abc.abstractmethod
    def fit(data):
        """Find the parameters necessary to trasform the data.

        Args:
            data ([type]): input data
        """
        pass

    @abc.abstractmethod
    def transform(data):
        """Transform the data.

        Args:
            data ([type]): input data
        """
        pass

    @abc.abstractmethod
    def fit_transform(data):
        """Combine :func:`~topicmodel.common.TopicModel.fit` and
        :func:`~topicmodel.common.TopicModel.transform`.

        This is to be preferred to calling fit and transform subsequently.

        Args:
            data ([type]): input data
        """
        pass



"""Common variable and function for data submodule."""

import abc


class TopicModel(abc.ABC):
    """Abstract class for topic modelling algorithms."""

    @abc.abstractmethod
    def fit(self, data):
        """Find all the relevant parameters to get
        the distribution of the topic among the documents.

        Args:
            data: input data, genrally a matrix of terms and documents.
        """
        pass

    @abc.abstractmethod
    def transform(self, data):
        """Retrieve the distribution of the topic among the documents from prevously
        computed parameters.

        Args:
            data: input data, genrally a matrix of terms and documents.
        """
        pass

    @abc.abstractmethod
    def fit_transform(self, data):
        """Retrieve the distribution of the topic among the documents, computing the
        parameters at the same time.

        This is to be preferred to calling :func:`~topicmodel.common.TopicModel.fit`
        and :func:`~topicmodel.common.TopicModel.transform` subsequently.

        Args:
            data: input data, genrally a matrix of terms and documents.
        """
        pass



"""Vectorizer class."""

from __future__ import annotations

from typing import Callable, List, Union

import nltk

from . import nltk_download
from .common import Transformer
from ..data import Dataset


def _lowercase(s: str) -> str:
    return s.lower()


class Tokenizer(Transformer):
    """Transform each document in a corpus in a list of tokens."""

    def __init__(
        self: Tokenizer,
        tokenizer: Callable[[str], List[str]] = nltk.tokenize.word_tokenize,
        stemmer: Union[None, Callable[[str], str]] = _lowercase,
        stopwords: Union[None, List[str]] = None,
    ) -> None:
        """Initialize the Tokenizer, specifing callable and parameters to be used.

        Args:
            tokenizer (callable, optional): any function which maps a string into
                a list of strings. Defaults to :func:`nltk.tokenize.word_tokenize`.
            stemmer (Union[None, callable], optional): any function which maps a string
                into a string, for example a stemmer. Applied after the tokenizer.
                Defaults to _lowercase.
            stopwords (Union[None, List[str]], optional): list of string to be ignored.
                Removed after calling the stemmer. Defaults to None.
        """
        if tokenizer == nltk.tokenize.word_tokenize:
            nltk_download("punkt")
        self.tokenizer = tokenizer

        self.stemmer = stemmer

        self.stopwords = stopwords

        if self.stemmer and self.stopwords:
            self.transformer = lambda s: list(
                filter(
                    lambda tkn: tkn not in self.stopwords,
                    map(self.stemmer, self.tokenizer(s)),
                )
            )
        elif self.stopwords:
            self.transformer = lambda s: list(
                filter(lambda tkn: tkn not in self.stopwords, self.tokenizer(s))
            )
        elif self.stemmer:
            self.transformer = lambda s: list(map(self.stemmer, self.tokenizer(s)))
        else:
            self.transformer = self.tokenizer

    def fit(self: Tokenizer, data: Dataset) -> None:
        """Do nothing method."""
        pass

    def transform(self: Tokenizer, data: Dataset) -> List[List[str]]:
        """Tokenize a given Dataset.

        Args:
            data (Dataset): a :class:`data.Dataset` to be tokenized

        Returns:
            List[List[str]]: a list of list of tokens. Outer list map to documents
                while inner list map to words in each document.
        """
        return list(map(self.transformer, data.get_data()))

    def fit_transform(self: Tokenizer, data: Dataset) -> List[List[str]]:
        """Alias for :func:`~transform.Tokenizer.transform`.

        Args:
            data (Dataset): a :class:`data.Dataset` to be tokenized

        Returns:
            List[List[str]]: a list of list of tokens. Outer list map to documents
                while inner list map to words in each document.
        """
        return self.transform(data)



"""Datasets and datased related functions."""

from __future__ import annotations

from . import newsgroups, test
from .common import DATADIR, load_txt
from .dataset import Dataset, LabelledDataset
from .newsgroups import load_20newsgroups
from .test import load_labelled_test_dataset, load_test_dataset

# Create data folder if it not exists
DATADIR.mkdir(exist_ok=True, parents=True)



"""Common variable and function for data submodule."""

import os
from itertools import chain
from pathlib import Path
from typing import Union


from .dataset import Dataset, LabelledDataset
from ..common import ADSODIR

DATADIR = ADSODIR / "data"


def load_txt(
    path: Union[str, bytes, os.PathLike],
    lines: bool = False,
    label: bool = False,
    extension: Union[str, None] = "txt",
    encoding: str = "utf-8",
    ignore_errors: bool = True,
) -> Dataset:
    """Load text files as dataset.

    Load a dataset composed by text files.
    This function scan the given directory and all its subdirectories and read all the
    files in them. The single items for the dataset could be either the files or each
    lines in each files. In the first case the folder's name will be used as label,
    in the latter the file's name.

    Args:
        path (Union[str, bytes, os.PathLike]): the directory to be scanned for the
            datatset.
        lines (bool, optional): if each item of the datatset is a different file (False)
            or each line in it (True). Defaults to False.
        label (bool, optional): to load the files as a Dataset (False) or a
            LabelledDataset, i.e. a dataset with labelled data (True). The folder names
            (if lines=False) or the file names (if lines=False) are used as labels.
            Defaults to False.
        extension (Union[str, None], optional): the extension of the file to be read.
            If None read all files. Defaults to "txt".
        encoding (str, optional): encoding for the files, passed to
            pathlib.Path.read_text(). Defaults to "utf-8".
        ignore_errors (bool, optional): whenever ignore unreadable characters in files
            or raise an error. Defaults to True.

    Returns:
        Dataset: return a Dataset, as defined in dataset.py file
    """
    path = Path(path)
    if extension is None:
        extension = ""
    else:
        extension = "." + extension

    if ignore_errors:
        errors = "ignore"
    else:
        errors = None

    if label:
        if lines:
            return LabelledDataset(
                list(
                    chain.from_iterable(
                        map(
                            lambda f: map(
                                lambda line: (line, f.stem),
                                f.read_text(
                                    encoding=encoding, errors=errors
                                ).splitlines(),
                            ),
                            filter(
                                lambda f: f.is_file(), path.glob("**/*" + extension)
                            ),
                        )
                    )
                )
            )
        else:  # label and not lines
            return LabelledDataset(
                list(
                    map(
                        lambda f: (
                            f.read_text(encoding=encoding, errors=errors),
                            f.parent.name,
                        ),
                        filter(lambda f: f.is_file(), path.glob("**/*" + extension)),
                    )
                )
            )
    else:  # not label
        if lines:
            return Dataset(
                list(
                    chain.from_iterable(
                        map(
                            lambda f: f.read_text(
                                encoding=encoding, errors=errors
                            ).splitlines(),
                            filter(
                                lambda f: f.is_file(), path.glob("**/*" + extension)
                            ),
                        )
                    )
                )
            )
        else:  # not label and not lines
            return Dataset(
                list(
                    map(
                        lambda f: f.read_text(encoding=encoding, errors=errors),
                        filter(lambda f: f.is_file(), path.glob("**/*" + extension)),
                    )
                )
            )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['adso', 'adso.data', 'adso.topicmodel', 'adso.transform']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.3.3,<4.0.0',
 'nltk>=3.5,<4.0',
 'pathlib>=1.0.1,<2.0.0',
 'requests>=2.25.0,<3.0.0',
 'scipy>=1.5.4,<2.0.0',
 'sparse>=0.11.2,<0.12.0']

setup_kwargs = {
    'name': 'adso',
    'version': '0.0.1.post1',
    'description': 'A topic-modelling library',
    'long_description': 'adso\n====\n\n**A topic modelling library built on top of scipy/numpy and nltk.**\n\ninstall\n^^^^^^^\n\nTo install::\n\n    pip install adso\n\nconfig\n^^^^^^\n\nadso need to write some files to disk.\nAs default adso uses the ``~/.adso`` folder, but it can be change setting the enviromental variable ``ADSODIR``.\n\ndocumentation\n^^^^^^^^^^^^^\n\nDocumentation with examples is hosted on [GitHub Pages](https://tnto.github.io/adso/index.html)\n\nSome examples on how to use adso are also available in ``tests`` and ``examples`` folders.\n\n\n',
    'author': "Michele 'TnTo' Ciruzzi",
    'author_email': 'tnto@hotmail.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/TnTo/adso',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.9',
}


setup(**setup_kwargs)



"""Dataset class.

Define data-container for other classes.
"""

from __future__ import annotations

from typing import List, Tuple


class Dataset:
    """Dataset class."""

    def __init__(self: Dataset, data: List[str]) -> None:
        """Constructor for Dataset class.

        Args:
            data (List[str]): A list of string, each will be considered as one document.
        """
        self.data: List[str] = data

    def get_data(self: Dataset) -> List[str]:
        """Access data stored into the dataset.

        Returns:
            List[str]: Return a list of string, each one is a document
        """
        return self.data

    def __len__(self: Dataset) -> int:
        """Return the number of the document in the dataset.

        Returns:
            int: number of document in the dataset
        """
        return len(self.data)

    def __getitem__(self: Dataset, index: int) -> str:
        """Enable the data[i] syntax to retrieve the data.

        Args:
            index (int): index of the required document

        Returns:
            str: the required document as string
        """
        return self.data[index]

    def __add__(self: Dataset, other: Dataset) -> Dataset:
        """Overload + operator to concatenate dataset.

        Args:
            other (Dataset): the second dataset to be concatenated

        Returns:
            Dataset: a dataset with the documents of both summed datasets
        """
        return Dataset(self.get_data() + other.get_data())


class LabelledDataset(Dataset):
    """Labbelled Dataset class.

    A subclass of :class:`Dataset` wich store also labels for documents.
    Documents and labels are stored in two different lists where each pair share
    the index.
    """

    def __init__(self: LabelledDataset, data: List[Tuple[str, str]]) -> None:
        """Contructor for LabelledDataset class.

        Args:
            data (List[Tuple[str, str]]): list of (document, label) tuples of strings.
        """
        self.data, self.labels = zip(*data)

    def get_data(self: LabelledDataset) -> List[str]:
        """Access data stored into the dataset.

        Returns:
            List[str]: Return a list of string, each one is a document
        """
        return self.data

    def get_labels(self: LabelledDataset) -> List[str]:
        """Access labels stored into the dataset.

        Returns:
            List[str]: Return a list of string, each one is a label
        """
        return self.labels

    def toDataset(self: LabelledDataset) -> Dataset:
        """Convert a labelledDataset to Dataset.

        Returns:
            Dataset: a Dataset with the same documents list but without labels.
        """
        return Dataset(self.get_data())

    def __getitem__(self: LabelledDataset, index: int) -> Tuple[str, str]:
        """Enable the data[i] syntax to retrieve the data.

        Args:
            index (int): index of the required document

        Returns:
            Tuple[str, str]: the required document and its label, as tuple
        """
        return self.data[index], self.labels[index]

    def __add__(self: LabelledDataset, other: Dataset) -> Dataset:
        """Overload + operator to concatenate dataset.

        Args:
            other (Dataset): the second dataset to be concatenated
                (must be a Dataset subclass)

        Returns:
            Dataset: a dataset with the documents of both summed dataset.
                If both Dataset are Labelled, instead, return a LabelledDataset
                with also the labels.
        """
        if isinstance(other, LabelledDataset):
            return LabelledDataset(
                list(zip(self.get_data(), self.get_labels()))
                + list(zip(other.get_data(), other.get_labels()))
            )
        else:
            return Dataset(self.get_data() + other.get_data())



"""Latent Dirichlet Allocation.

From Blei et al. 2003.
"""

from __future__ import annotations

import warnings
from typing import Tuple

import numpy as np
import scipy as sp
import sparse

from .common import TopicModel


class LDA(TopicModel):
    """Latent Dirichlet Allocation.

    Implemented following Blei, Jordan, Ng 2003.
    """

    def __init__(
        self: LDA,
        n_topic: int = 10,
        max_iter: int = 200,
        tolerance: float = 0,
        epsilon: float = 1e-50,
    ) -> None:
        """Initialize the algorithm.

        Args:
            n_topic (int, optional): number of topics in copus to be estimated.
                Defaults to 10.
            max_iter (int, optional): max number of iterations for iterative steps.
                Defaults to 200.
            tolerance (float, optional): target relative error for early stopping
                in iterative steps. Defaults to 0.
            epsilon (float, optional): small number to avoid -inf in logarithm result.
                Defaults to 1e-50.
        """
        if n_topic > 0:
            self.n_topic = n_topic
        else:
            raise ValueError("n_topic must be positive")
        if max_iter > 0:
            self.max_iter = max_iter
        else:
            raise ValueError("max_iter must be positive")
        if tolerance >= 0:
            self.tolerance = tolerance
        else:
            raise ValueError("tolerance must be non-negative")
        if 1e-10 > epsilon > 0:
            self.epsilon = epsilon
        else:
            raise ValueError("epsilon must be positive and small")

    # data count documwnt-term matrix
    def fit(
        self: LDA, data: sp.sparse.spmatrix
    ) -> Tuple[np.array, np.array, np.array, np.array, float, int]:
        """Estimate LDA parameters, particularly alpha e beta.

        Args:
            data (sp.sparse.spmatrix): document-term matrix with counts as entry.
                For example the output of :class:`CountVectorizer`

        Returns:
            Tuple[np.array, np.array, np.array, np.array, float, int]:
                alpha (np.array):
                    Dirichlet priors for LDA model (also stored as instance attributes)
                beta (np.array):
                    Multinomial priors for LDA model (also stored as instance
                    attributes)
                gamma (np.array):
                    Dirichlet priors for auxiliary model
                phi (np.array):
                    Multinomial priors for auxiliary model
                likelihood (float):
                    achieved likelihood
                iter (int):
                    number of iteration of EM algorithm
        """
        n_topic = self.n_topic
        n_doc, n_term = data.shape

        data = sparse.COO.from_scipy_sparse(data)

        if n_doc / n_topic > 5:
            p = 5
        else:
            p = max(int(n_doc / n_topic), 1)

        alpha = np.ones(n_topic)
        beta = (
            data[np.random.choice(n_doc, size=p * n_topic, replace=False), :]
            .reshape((n_topic, n_term, p))
            .sum(axis=2)
        )
        beta = beta / beta.sum(axis=1, keepdims=True).todense()

        phi = sparse.full((n_doc, n_term, n_topic), 1 / n_topic)
        gamma = (
            alpha[np.newaxis, :] + data.sum(axis=1, keepdims=True).todense() / n_topic
        )

        loglikelihood = self._loglikelihood(data, alpha, beta, gamma, phi)
        iter: int = 0
        error: float = 1

        while iter < self.max_iter and error > self.tolerance:

            loglikelihood_old = loglikelihood

            gamma, phi = self._estep(alpha, beta, gamma)
            alpha, beta, _ = self._mstep(data, alpha, gamma, phi)

            loglikelihood = self._loglikelihood(data, alpha, beta, gamma, phi)

            error = abs((loglikelihood_old - loglikelihood) / loglikelihood_old)
            iter += 1
            print(f"Iteration {iter} Log-Likelihood {loglikelihood}")

        beta = beta.tocsr()

        self.alpha = alpha
        self.beta = beta

        return alpha, beta, gamma, phi, loglikelihood, iter

    def transform(self: LDA, data: sp.sparse.spmatrix) -> sp.sparse.spmatrix:
        """Estimate the probability of each topic in each document, given the model.

        Must be called after a :func:`~topicmodel.LDA.fit` method on the same instance.

        Args:
            data (sp.sparse.spmatrix): document-term matrix with counts as entry.
                For example the output of :class:`CountVectorizer`

        Returns:
            sp.sparse.spmatrix: the probability of each document (rows) given the
                different topics (columns).
        """
        return sp.sparse.csr_matrix(
            np.exp(data @ np.log(self.beta.T.todense() + self.epsilon))
        )

    def fit_transform(
        self: LDA, data: sp.sparse.spmatrix
    ) -> Tuple[sp.sparse.matrix, np.array, np.array, np.array, np.array, float, int]:
        """Estimate the probability of each topic in each document.

        Perform the estimation of both model's parameters (:func:`~topicmodel.LDA.fit`)
        and topic-document relation (:func:`~topicmodel.LDA.transform`).

        Args:
            data (sp.sparse.spmatrix): document-term matrix with counts as entry.
                For example the output of :class:`CountVectorizer`

        Returns:
            Tuple[sp.sparse.matrix, np.array, np.array, np.array, np.array, float, int]:
                estimation (sp.sparse.spmatrix):
                    the probability of each document (rows) given the different topics
                    (columns).
                alpha (np.array):
                    Dirichlet priors for LDA model (also stored as instance attributes)
                beta (np.array):
                    Multinomial priors for LDA model (also stored as instance
                    attributes)
                gamma (np.array):
                    Dirichlet priors for auxiliary model
                phi (np.array):
                    Multinomial priors for auxiliary model
                likelihood (float):
                    achieved likelihood
                iter (int):
                    number of iteration of EM algorithm
        """
        alpha, beta, gamma, phi, loglikelihood, iter = self.fit(data)
        return self.transform(data), alpha, beta, gamma, phi, loglikelihood, iter

    def _loglikelihood(
        self: LDA,
        data: sp.sparse.matrix,  # dw
        alpha: np.array,  # i
        beta: np.array,  # iw
        gamma: np.array,  # di
        phi: np.array,  # dwi
    ) -> float:
        """Estimate a lower bound for the log-likelihood, using Jensen's inequality
        and auxiliary model.

        Args:
            data (sp.sparse.matrix): document-term matrix with counts as entry.
                For example the output of :class:`CountVectorizer`

        Returns:
            float: the computed lower bound for the loglikelihood.
        """
        n_doc, n_term = data.shape
        sum_digamma = sp.special.digamma(gamma) - sp.special.digamma(
            gamma.sum(axis=1, keepdims=True)
        )  # di
        likelihood = sp.special.loggamma(alpha.sum()) * n_doc
        likelihood -= sp.special.loggamma(alpha).sum() * n_doc
        likelihood += ((alpha - 1) * (sum_digamma.sum(axis=0))).sum()
        likelihood += (phi.sum(axis=1) * sum_digamma).sum()
        likelihood += (
            phi
            * sparse.elemwise(np.nan_to_num, np.log(beta + self.epsilon))[
                :, :, np.newaxis
            ].T
            * data[:, :, np.newaxis]
        ).sum()
        likelihood -= sp.special.loggamma(gamma.sum(axis=1)).sum()
        likelihood += sp.special.loggamma(gamma).sum()
        likelihood -= ((gamma - 1) * sum_digamma).sum()
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", r".*((log)|(multiply)).*", category=RuntimeWarning
            )
            likelihood -= sparse.elemwise(np.nan_to_num, phi * np.log(phi)).sum()
        return likelihood

    def _estep(
        self: LDA,
        alpha: np.array,
        beta: np.array,
        gamma: np.array,
    ) -> Tuple[np.array, np.array]:
        """Estimating the auxiliary model parameters to minimize the KL-divergence
        between the auxiliary model and the last estimation of LDA model given
        the data.

        Args:
            alpha (np.array): estimation of Dirichlet parameters for the LDA model
            beta (np.array): estimation of multinomial parameters for the LDA model
            gamma (np.array): last estimation of Dirichlet parameters for the
                auxiliary model

        Returns:
            Tuple[np.array, np.array]: gamma (Dirichlet parameters) and phi
                (multinomial parameters) for auxiliary model.
                Note that gamma and phi are document specific.
        """
        phi = (
            np.exp(sp.special.digamma(gamma))[:, np.newaxis, :]
            * beta[:, :, np.newaxis].T
        )
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", r".*(true_divide).*", category=RuntimeWarning
            )
            phi = sparse.elemwise(np.nan_to_num, phi / phi.sum(axis=2, keepdims=True))
        gamma = alpha[np.newaxis, :] + phi.sum(axis=1, keepdims=False).todense()

        return gamma, phi

    def _mstep(
        self: LDA,
        data: sp.sparse.matrix,
        alpha: np.array,
        gamma: np.array,
        phi: np.array,
    ) -> Tuple[np.array, np.array, float]:
        """Estimating the LDA parameters with maximum likelihood estimation based
        on data and approximate posterior distribution given by the auxiliary model.

        Args:
            data (sp.sparse.matrix): document-term matrix with counts as entry.
                For example the output of :class:`CountVectorizer`
            alpha (np.array): last estimation of Dirichlet parameters for the LDA model
            gamma (np.array): estimation of Dirichlet parameters for the
                auxiliary model
            phi (np.array): estimation of multinomial parameters for the
                auxiliary model

        Returns:
            Tuple[np.array, np.array, int]: alpha (Dirichlet parameters) and beta
                (multinomial parameters) for LDA model and the number of iterations
                computed for the estimation of alpha with a Newton-like method.
        """
        beta = (phi * data[:, :, np.newaxis]).sum(axis=0).T
        beta = beta / beta.sum(axis=1, keepdims=True).todense()
        alpha, iter_alpha = self._update_alpha(alpha, gamma)

        return alpha, beta, iter_alpha

    def _update_alpha(
        self: LDA, alpha: np.array, gamma: np.array
    ) -> Tuple[np.array, int]:
        """Modified Newton-Raphson method to update Diriclet priors for LDA.

        Args:
            alpha (np.array): inital alpha
            gamma (np.array): parameters for Dirichlet priors of auxiliary model

        Returns:
            Tuple[np.array, int]: return updated alpha and the number of iteration.
        """
        iter = 0
        error = 1
        init_alpha = alpha
        n_doc = gamma.shape[0]
        while iter < self.max_iter and error > self.tolerance and (alpha > 0).all():
            alpha_old = alpha
            g = -n_doc * (
                sp.special.digamma(alpha_old) - sp.special.digamma(alpha_old.sum())
            ) + (
                sp.special.digamma(gamma)
                - sp.special.digamma(gamma.sum(axis=1, keepdims=True))
            ).sum(
                axis=0
            )
            h = -n_doc * sp.special.polygamma(1, alpha)
            z = sp.special.polygamma(1, alpha.sum())
            c = (g / h).sum() / (np.reciprocal(z) + np.reciprocal(h).sum())
            alpha = alpha_old - (g - c) / h

            iter += 1
            error = (np.abs(alpha_old - alpha) / np.abs(alpha_old)).sum()

        if np.isnan(alpha).any() or (alpha <= 0).any():
            self._update_alpha(init_alpha * 2 + 1e-3, gamma)

        return alpha, iter
